@@ -1,16 +1,11 @@
 from ssl import SSLContext
-from typing import Optional, Tuple, cast
+from typing import List, Optional, Tuple, cast
 
 from .._backends.auto import AsyncBackend, AsyncLock, AsyncSocketStream, AutoBackend
 from .._exceptions import ConnectError, ConnectTimeout
 from .._types import URL, Headers, Origin, TimeoutDict
 from .._utils import exponential_backoff, get_logger, url_to_origin
-from .base import (
-    AsyncByteStream,
-    AsyncHTTPTransport,
-    ConnectionState,
-    NewConnectionRequired,
-)
+from .base import AsyncByteStream, AsyncHTTPTransport, NewConnectionRequired
 from .http import AsyncBaseHTTPConnection
 from .http11 import AsyncHTTP11Connection
 
@@ -23,7 +18,9 @@ class AsyncHTTPConnection(AsyncHTTPTransport):
     def __init__(
         self,
         origin: Origin,
+        http1: bool = True,
         http2: bool = False,
+        keepalive_expiry: float = None,
         uds: str = None,
         ssl_context: SSLContext = None,
         socket: AsyncSocketStream = None,
@@ -32,15 +29,22 @@ class AsyncHTTPConnection(AsyncHTTPTransport):
         backend: AsyncBackend = None,
     ):
         self.origin = origin
+        self.http1 = http1
         self.http2 = http2
+        self.keepalive_expiry = keepalive_expiry
         self.uds = uds
         self.ssl_context = SSLContext() if ssl_context is None else ssl_context
         self.socket = socket
         self.local_address = local_address
         self.retries = retries
 
-        if self.http2:
-            self.ssl_context.set_alpn_protocols(["http/1.1", "h2"])
+        alpn_protocols: List[str] = []
+        if http1:
+            alpn_protocols.append("http/1.1")
+        if http2:
+            alpn_protocols.append("h2")
+
+        self.ssl_context.set_alpn_protocols(alpn_protocols)
 
         self.connection: Optional[AsyncBaseHTTPConnection] = None
         self.is_http11 = False
@@ -50,20 +54,58 @@ class AsyncHTTPConnection(AsyncHTTPTransport):
         self.backend = AutoBackend() if backend is None else backend
 
     def __repr__(self) -> str:
-        http_version = "UNKNOWN"
-        if self.is_http11:
-            http_version = "HTTP/1.1"
-        elif self.is_http2:
-            http_version = "HTTP/2"
-        return f"<AsyncHTTPConnection http_version={http_version} state={self.state}>"
+        return f"<AsyncHTTPConnection [{self.info()}]>"
 
     def info(self) -> str:
         if self.connection is None:
-            return "Not connected"
-        elif self.state == ConnectionState.PENDING:
-            return "Connecting"
+            return "Connection failed" if self.connect_failed else "Connecting"
         return self.connection.info()
 
+    def should_close(self) -> bool:
+        """
+        Return `True` if the connection is in a state where it should be closed.
+        This occurs when any of the following occur:
+
+        * There are no active requests on an HTTP/1.1 connection, and the underlying
+          socket is readable. The only valid state the socket can be readable in
+          if this occurs is when the b"" EOF marker is about to be returned,
+          indicating a server disconnect.
+        * There are no active requests being made and the keepalive timeout has passed.
+        """
+        if self.connection is None:
+            return False
+        return self.connection.should_close()
+
+    def is_idle(self) -> bool:
+        """
+        Return `True` if the connection is currently idle.
+        """
+        if self.connection is None:
+            return False
+        return self.connection.is_idle()
+
+    def is_closed(self) -> bool:
+        if self.connection is None:
+            return self.connect_failed
+        return self.connection.is_closed()
+
+    def is_available(self) -> bool:
+        """
+        Return `True` if the connection is currently able to accept an outgoing request.
+        This occurs when any of the following occur:
+
+        * The connection has not yet been opened, and HTTP/2 support is enabled.
+          We don't *know* at this point if we'll end up on an HTTP/2 connection or
+          not, but we *might* do, so we indicate availability.
+        * The connection has been opened, and is currently idle.
+        * The connection is open, and is an HTTP/2 connection. The connection must
+          also not currently be exceeding the maximum number of allowable concurrent
+          streams and must not have exhausted the maximum total number of stream IDs.
+        """
+        if self.connection is None:
+            return self.http2 and not self.is_closed
+        return self.connection.is_available()
+
     @property
     def request_lock(self) -> AsyncLock:
         # We do this lazily, to make sure backend autodetection always
@@ -84,18 +126,16 @@ class AsyncHTTPConnection(AsyncHTTPTransport):
         timeout = cast(TimeoutDict, extensions.get("timeout", {}))
 
         async with self.request_lock:
-            if self.state == ConnectionState.PENDING:
+            if self.connection is None:
+                if self.connect_failed:
+                    raise NewConnectionRequired()
                 if not self.socket:
                     logger.trace(
                         "open_socket origin=%r timeout=%r", self.origin, timeout
                     )
                     self.socket = await self._open_socket(timeout)
                 self._create_connection(self.socket)
-            elif self.state in (ConnectionState.READY, ConnectionState.IDLE):
-                pass
-            elif self.state == ConnectionState.ACTIVE and self.is_http2:
-                pass
-            else:
+            elif not self.connection.is_available():
                 raise NewConnectionRequired()
 
         assert self.connection is not None
@@ -147,38 +187,29 @@ class AsyncHTTPConnection(AsyncHTTPTransport):
         logger.trace(
             "create_connection socket=%r http_version=%r", socket, http_version
         )
-        if http_version == "HTTP/2":
+        if http_version == "HTTP/2" or (self.http2 and not self.http1):
             from .http2 import AsyncHTTP2Connection
 
             self.is_http2 = True
             self.connection = AsyncHTTP2Connection(
-                socket=socket, backend=self.backend, ssl_context=self.ssl_context
+                socket=socket,
+                keepalive_expiry=self.keepalive_expiry,
+                backend=self.backend,
             )
         else:
             self.is_http11 = True
             self.connection = AsyncHTTP11Connection(
-                socket=socket, ssl_context=self.ssl_context
+                socket=socket, keepalive_expiry=self.keepalive_expiry
             )
 
-    @property
-    def state(self) -> ConnectionState:
-        if self.connect_failed:
-            return ConnectionState.CLOSED
-        elif self.connection is None:
-            return ConnectionState.PENDING
-        return self.connection.get_state()
-
-    def is_socket_readable(self) -> bool:
-        return self.connection is not None and self.connection.is_socket_readable()
-
-    def mark_as_ready(self) -> None:
-        if self.connection is not None:
-            self.connection.mark_as_ready()
-
-    async def start_tls(self, hostname: bytes, timeout: TimeoutDict = None) -> None:
+    async def start_tls(
+        self, hostname: bytes, ssl_context: SSLContext, timeout: TimeoutDict = None
+    ) -> None:
         if self.connection is not None:
             logger.trace("start_tls hostname=%r timeout=%r", hostname, timeout)
-            self.socket = await self.connection.start_tls(hostname, timeout)
+            self.socket = await self.connection.start_tls(
+                hostname, ssl_context, timeout
+            )
             logger.trace("start_tls complete hostname=%r timeout=%r", hostname, timeout)
 
     async def aclose(self) -> None:
